from .backend import get_backend
from .common import Color, Shape
import random
import math
from .config import Config

name = "sketcher"

config = Config()


class Sketch:
    def __init__(self):
        if not hasattr(self, 'Backend'):
            self.Backend = get_backend()

    def start(self):
        self.can = self.Backend()
        self.can.init()
        self.can.start(self.setup, self.loop)

    def clear(self):
        self.can.clear()

    def frame(self, l):
        self.can.set_frame(l)

    def fill(self):
        self.can.set_fill(True)

    def no_fill(self):
        self.can.set_fill(False)

    def stroke(self):
        self.can.set_stroke(True)

    def no_stroke(self):
        self.can.set_stroke(False)

    def fill_color(self, col):
        self.can.set_fill_color(col)

    def stroke_color(self, col):
        self.can.set_stroke_color(col)

    def mouse_state(self):
        return self.can.get_mouse_state()

    def keyboard_state(self, ):
        return self.can.get_keyboard_state()

    def size(self, w, h):
        self.can.set_size(w, h)

    def background(self, color):
        self.can.set_background(Color(color))

    def shape(self, shape):
        if not isinstance(shape, Shape):
            shape = Shape(list(shape))
        self.can.draw_shape(shape)

    def point(self, x, y):
        self.can.draw_point(x, y)

    def line(self, x1, y1, x2, y2):
        self.can.draw_line(x1, y1, x2, y2)

    def rectangle(self, x, y, w, h=None):
        if h is None:
            h = w
        self.can.draw_rectangle(x, y, w, h)

    def ellipse(self, x, y, a, b=None):
        if b is None:
            b = a
        self.can.draw_ellipse(x, y, a, b)

    def text(self, x, y, text, **kwargs):
        self.can.draw_text(x, y, text, **kwargs)

    def image(self, x, y, image):
        self.can.draw_image(x, y, image)

    def refresh(self):
        self.can.refesh()


Sketch.m = math
Sketch.rd = random

try:
    import numpy
    Sketch.np = numpy
except ImportError:
    Sketch.np = None


def sketch(sk):
    if issubclass(sk, Sketch):
        MySketch = sk
    else:
        class MySketch(Sketch, sk):
            pass
    MySketch.Backend = get_backend(config.backend_use)
    MySketch.__name__ = sk.__name__
    MySketch().start()
    return MySketch



"""
This just print an empty sketch to help user starting.
"""
empty = """
#!/usr/bin/env python
from sketcher import sketch, Sketch


@sketch
class Sk(Sketch):
    # you do not need to inherit from sketch but it help auto-completion
    def setup(self):
        pass  # Here come the initialisation

    def loop(self):
        pass  # Here come the per-frame actions
"""
print(empty)



from .backend_base import CanvasBackend
from .common import KeyboardState, MouseState, Color
import tkinter as tk
from tkinter import font as tkfont
from queue import Queue
try:
    from PIL import Image, ImageTk

    def load_image(filename):
        return ImageTk.PhotoImage(Image.open(filename))

except ImportError:
    print('PIL not available. Only GIF and PGM images can be opened.')
    from tkinter import PhotoImage

    def load_image(filename):
        return PhotoImage(filename)

class Backend(CanvasBackend):
    def __init__(self):
        CanvasBackend.__init__(self)
        self.win = tk.Tk()
        self.can = tk.Canvas(self.win)
        self.event_queue = Queue()

        self.stroke_color = Color('black')
        self.fill_color = Color('red')
        self.back_color = Color('white')
        self.image_cache = {}

        def add_event(type):
            def adder(ev):
                self.event_queue.put((type, ev))
            return adder

        self.can.bind('<KeyPress>', add_event('key_press'))
        self.can.bind('<KeyRelease>', add_event('key_release'))
        self.can.bind('<ButtonPress>', add_event('mouse_press'))
        self.can.bind('<ButtonRelease>', add_event('mouse_release'))
        self.can.bind('<Motion>', add_event('mouse_move'))

        self.user_loop = None

        self.__keyboard_state = KeyboardState()
        self.__mouse_state = MouseState()

        self.__pix = tk.PhotoImage(width=1, height=1)

    def init(self):
        self.can.configure(
            width=self.size[0],
            height=self.size[1],
            background=self.back_color.hashtag()
        )
        self.can.pack()
        self.win.focus_set()

    def start(self, setup, loop):
        setup()
        self.user_loop = loop
        self.win.after(int(1000*self.frame), self.loop)
        self.win.mainloop()

    def loop(self):
        self.win.after(int(1000*self.frame), self.loop)
        self.can.focus_set()

        # poll event
        self.__mouse_state.clean()
        self.__keyboard_state.clean()

        while not self.event_queue.empty():
            ev_type, ev = self.event_queue.get()
            if ev_type == 'key_press':
                if ev.keysym:
                    self.__keyboard_state.pressed.add(ev.keysym)
                else:
                    print(ev.keycode, ev.keysym)
                    raise NotImplementedError
            elif ev_type == 'key_release':
                if ev.keysym:
                    self.__keyboard_state.released.add(ev.keysym)
                else:
                    print(ev.keycode, ev.keysym)
                    raise NotImplementedError
            elif ev_type == 'mouse_move':
                self.__mouse_state.pos = (ev.x, ev.y)

            elif ev_type == 'mouse_press':
                self.__mouse_state.pressed.add(ev.num)
                self.__mouse_state.pos = (ev.x, ev.y)

            elif ev_type == 'mouse_release':
                self.__mouse_state.released.add(ev.num)
                self.__mouse_state.pos = (ev.x, ev.y)
        self.__mouse_state.clean()
        self.__keyboard_state.clean()

        self.user_loop()

    def clear(self):
        self.can.delete('all')

    def set_fill(self, yes):
        self.fill = yes

    def set_stroke(self, yes):
        self.stroke = yes

    def set_stroke_color(self, color):
        self.stroke_color = Color(color)

    def set_fill_color(self, color):
        self.fill_color = Color(color)

    def get_mouse_state(self):
        return self.__mouse_state

    def get_keyboard_state(self):
        return self.__keyboard_state

    def set_size(self, w, h):
        self.size = (w, h)
        self.can.configure(width=w, height=h)

    def set_background(self, color):
        self.back_color = color
        self.can.configure(background=color.hashtag())

    def draw_point(self, x, y):
        fill = self.stroke_color.hashtag() if self.stroke else ''
        self.__pix.put(fill)
        self.can.create_image((x, y), image=self.__pix)

    def draw_line(self, x1, y1, x2, y2):
        fill = self.stroke_color.hashtag() if self.stroke else ''
        self.can.create_line(x1, y1, x2, y2, fill=fill)

    def draw_rectangle(self, x, y, w, h):
        fill = self.fill_color.hashtag() if self.fill else ''
        outline = self.stroke_color.hashtag() if self.stroke else ''
        self.can.create_rectangle(x, y, x+w, y+h, fill=fill, outline=outline)

    def draw_ellipse(self, x, y, a, b):
        x0 = x - a/2
        x1 = x + a/2
        y0 = y - b/2
        y1 = y + b/2
        fill = self.fill_color.hashtag() if self.fill else ''
        outline = self.stroke_color.hashtag() if self.stroke else ''
        self.can.create_oval(x0, y0, x1, y1, fill=fill, outline=outline)

    def draw_text(self, x, y, text, **kwargs):
        if 'font' in kwargs:
            font = kwargs['font']
        else:
            font = tkfont.Font(self.win)
        self.can.create_text(x, y, text=text, font=font)

    def draw_shape(self, shape):
        lst = []
        for x, y in shape.vertex:
            lst.append(x)
            lst.append(y)
        fill = self.fill_color.hashtag() if self.fill else ''
        outline = self.stroke_color.hashtag() if self.stroke else ''
        self.can.create_polygon(*lst, fill=fill, outline=outline)

    def draw_image(self, x, y, filename, scale=(1, 1)):
        if filename not in self.image_cache:
            self.image_cache[filename] = load_image(filename)
        self.can.create_image((x, y), image=self.image_cache[filename])



from .backend_base import CanvasBackend
from .common import KeyboardState, MouseState, Color
import pyglet as pg
from queue import Queue


def mouse_button(b):
    if b == pg.window.mouse.LEFT:
        return 1
    elif b == pg.window.mouse.RIGHT:
        return 3
    elif b == pg.window.mouse.MIDDLE:
        return 2
    else:
        return 0


def key_symbol(b):
    return pg.window.key.symbol_string(b)


class Backend(CanvasBackend):
    def __init__(self):
        CanvasBackend.__init__(self)
        self.win = pg.window.Window(vsync=0)
        self.event_queue = Queue()

        self.stroke_color = Color('black')
        self.fill_color = Color('red')
        self.back_color = Color('white')
        self.batch = pg.graphics.Batch()

        @self.win.event
        def on_key_press(symbol, modifiers):
            self.event_queue.put(('key_press', (symbol, modifiers)))

        @self.win.event
        def on_key_release(symbol, modifiers):
            self.event_queue.put(('key_release', (symbol, modifiers)))

        @self.win.event
        def on_mouse_press(x, y, button, modifiers):
            self.event_queue.put(('mouse_press', (x, y, button, modifiers)))

        @self.win.event
        def on_mouse_release(x, y, button, modifiers):
            self.event_queue.put(('mouse_release', (x, y, button, modifiers)))

        @self.win.event
        def on_mouse_motion(x, y, dx, dy):
            self.event_queue.put(('mouse_move', (x, y)))

        self.user_loop = None

        self.__keyboard_state = KeyboardState()
        self.__mouse_state = MouseState()

    def init(self):
        pass
        # setup background, canvas, focus

    def start(self, setup, loop):
        setup()
        self.user_loop = loop

        # run main loop
        self.loop()
        pg.app.run()

    def loop(self, dt=0):
        pg.clock.schedule_once(self.loop, self.frame)
        # poll event
        self.__mouse_state.clean()
        self.__keyboard_state.clean()

        while not self.event_queue.empty():
            # empty event list
            ev_type, ev = self.event_queue.get()
            if ev_type == 'key_press':
                self.__keyboard_state.pressed.add(key_symbol(ev[0]))
            elif ev_type == 'key_release':
                self.__keyboard_state.pressed.add(key_symbol(ev[0]))
            elif ev_type == 'mouse_move':
                self.__mouse_state.pos = (ev[0], self.size[1]-ev[1])

            elif ev_type == 'mouse_press':
                self.__mouse_state.pressed.add(mouse_button(ev[2]))
                self.__mouse_state.pos = (ev[0], self.size[1]-ev[1])

            elif ev_type == 'mouse_release':
                self.__mouse_state.pressed.add(mouse_button(ev[2]))
                self.__mouse_state.pos = (ev[0], self.size[1]-ev[1])

        self.__mouse_state.clean()
        self.__keyboard_state.clean()

        self.win.clear()
        self.user_loop()
        self.background.draw()
        self.batch.draw()

    def clear(self):
        # clear canvas
        self.win.clear()

    def set_fill(self, yes):
        self.fill = yes

    def set_stroke(self, yes):
        self.stroke = yes

    def set_stroke_color(self, color):
        self.stroke_color = Color(color)

    def set_fill_color(self, color):
        self.fill_color = Color(color)

    def get_mouse_state(self):
        return self.__mouse_state

    def get_keyboard_state(self):
        return self.__keyboard_state

    def set_size(self, w, h):
        self.size = (w, h)
        self.win.set_size(w, h)

    def set_background(self, color):
        self.back_color = color
        # set background color

    def draw_point(self, x, y):
        self.batch.add(1, pg.gl.GL_POINTS, None,
                       ('v2f', float(x), float(y)),
                       ('c3B', self.stroke_color.tupple255()))

    def draw_line(self, x1, y1, x2, y2):
        self.batch.add(2, pg.gl.GL_LINES, None,
                       ('v2f', (x1, y1, x2, y2)),
                       ('c3B', self.stroke_color.tupple255()*2))

    def draw_rectangle(self, x, y, w, h):
        pass

    def draw_ellipse(self, x, y, a, b):
        pass

    def draw_text(self, x, y, text, **kwargs):
        pass

    def draw_shape(self, shape):
        pass



class NoBackend(BaseException):
    def __init__(self):
        super().__init__(
            "There is no supported graphical backend installed.\n" +
            "Try to install one of those:\n" +
            "\n".join("- " + back for back in supported_backends)
        )


supported_backends = [
    # "pyglet",  # not ready yet
    "tkinter",
]

try:
    import pyglet
    have_pyglet = True
except ImportError:
    have_pyglet = False


try:
    import tkinter
    have_tk = True
except ImportError:
    have_tk = False
have_pyglet = False  # in progress


def get_backend(use='auto'):
    if use == 'auto':
        if have_pyglet:
            use = 'pyglet'
        elif have_tk:
            use = 'tk'
        else:
            use = 'nothing'

    if use == 'pyglet':
        from .backend_pyglet import Backend
    elif use == 'tk':
        from .backend_tk import Backend
    else:
        raise NoBackend()
    return Backend



from __future__ import division
import math
from math import sin, cos
color_tab = {
    'black': (0, 0, 0),
    'white': (255, 255, 255),
    'red': (255, 0, 0),
    'green': (0, 255, 0),
    'blue': (0, 0, 255),
    'grey': (127, 127, 127),
}


class Shape:
    def __init__(self, vertex):
        self.vertex = vertex

    def center(self):
        n = len(self.vertex)
        cx = sum(x for x, y in self.vertex)/n
        cy = sum(y for x, y in self.vertex)/n
        return cx, cy

    def rotate(self, angle, center='center'):
        newv = []
        if center == 'center':
            center = self.center()
        for x, y in self.vertex:
            dx, dy = x - center[0], y - center[1]
            dxp = dx * cos(angle) + dy * sin(angle)
            dyp = - dx * sin(angle) + dy * cos(angle)
            newv.append((center[0] + dxp, center[1] + dyp))
        return Shape(newv)

    def translate(self, vec):
        newv = []
        for i in range(len(self.vertex)):
            newv.append(self.vertex[i][0] + vec[0],
                        self.vertex[i][1] + vec[1])
        return Shape(newv)

    def copy(self):
        return Shape(self.vertex.copy())


class Color:
    def __init__(self, col):
        self.r = 0
        self.g = 0
        self.b = 0
        if isinstance(col, str):
            self.r, self.g, self.b = color_tab[col]
        elif isinstance(col, Color):
            self.r = col.r
            self.g = col.g
            self.b = col.b
        elif isinstance(col, float) or isinstance(col, int):
            if math.isnan(col):
                col = 255
            elif col < 0:
                col = 0
            elif col > 255:
                col = 255
            self.r = int(col)
            self.g = int(col)
            self.b = int(col)
        elif hasattr(col, '__iter__'):
            self.r, self.g, self.b = col
        else:
            print(type(col))
            exit()

    def hashtag(self):
        r = hex(int(self.r))[2:]
        g = hex(int(self.g))[2:]
        b = hex(int(self.b))[2:]
        if len(r) != 2:
            r = '0' + r
        if len(g) != 2:
            g = '0' + g
        if len(b) != 2:
            b = '0' + b
        return '#{}{}{}'.format(r, g, b)

    def tupple255(self):
        return (int(self.r), int(self.g), int(self.b))


class MouseState:
    def __init__(self):
        self.pos = (0, 0)
        self.pressed = set()
        self.released = set()

    def flush(self):
        self.pressed.clear()
        self.released.clear()

    def clean(self):
        still_pressed = self.pressed - self.released
        releasing = self.released.intersection(self.pressed)

        self.pressed = still_pressed
        self.released = releasing


class KeyboardState:
    def __init__(self):
        self.pressed = set()
        self.released = set()

    def flush(self):
        self.pressed.clear()
        self.released.clear()

    def clean(self):
        still_pressed = self.pressed - self.released
        releasing = self.released.intersection(self.pressed)

        self.pressed = still_pressed
        self.released = releasing
